

#include <iostream>
#include <iomanip>
#include <fstream>
//#include <cstdlib>
#include <ctime>



#include "pari.h"
#include "bessel.c"

using namespace std;



const double pi= 3.141592653589793238462643;
double T = 1e-6;


long prec;

//****** class of functions for integrating and root finding *************

class func {
  public:
    virtual double f(double)=0;
};

//including the differential for advaced root finding
class Dfunc : func {
  public:
    double Df;
    virtual double f(double)=0;
};


// ********************** Matching one state *****************************


class matchE : public func {
  protected:
    int m;
    double s;
  public:
    matchE(int m1, double s1) {m=m1;s=s1;}
    double f(double);
};

class matchDE : public Dfunc {
  private:
    int m;
    double s;
  public:
    matchDE(int m1, double s1) {m=m1;s=s1;Df=0;}
    double duE(double, double, double);
    double f(double);
};

class findWKB0 : public func {
  private:
    int m,n;
    double s;
  public:
    findWKB0(int n1, int m1, double s1) {n=n1;m=m1;s=s1;}
    double nu(double, double);
    double f(double);
};

class findWKB1 : public func {
  private:
    int m,n;
    double s;
  public:
    findWKB1(int n1, int m1, double s1) {n=n1;m=m1;s=s1;}
    void nu(double, double, double, double&, double&, double&);
    double f(double);
};


// ********************** Wave function of one state *****************************


class unnormalized : public func {
  private:
    int m;
    double s;
    double E;
    bool squared;
  public:
    unnormalized(int m1, double s1, double E1, bool sq) {m=m1;s=s1;E=E1;squared=sq;}
    unnormalized(int m1, double s1, double E1) {m=m1;s=s1;E=E1;squared=false;}
    double W(double,double,double);
    double f(double);
};


class current : public func {
  private:
    int m,n;
    double s, E, norm, rmax;
    char *currentfilename;
    char *logfilename;
    bool normalbasis;
  public:
    current(int n1, int m1, double s1, double E1, double norm1, double rmax1) 
	{n=n1;m=m1;s=s1;E=E1;norm=norm1;rmax=rmax1;  logfilename = new char[30]; strcpy(logfilename,"current.log"); normalbasis=false;}
    current(int n1, int m1, double s1, double E1, double norm1, double rmax1, bool nb1) 
	{n=n1;m=m1;s=s1;E=E1;norm=norm1;rmax=rmax1;  logfilename = new char[30]; strcpy(logfilename,"current.log");normalbasis=nb1;}

    current() {  logfilename = new char[30]; strcpy(logfilename,"current.log");}
    double f(double);
    void display(int,ostream&);
    void save(int);
    void writelog();
    void resetlog();
};

// *************************** Matching **********************************

class Matching {
  private:
    char *qmfilename,*wkbfilename,*wkbfilename1,*normfilename,*Ifilename,*rmaxfilename,*logfilename;
    double rootfinder(double,double,int&,int);
    void BuildExact();
    void BuildWKB();
    void BuildWKB1();
    void BuildNorm();
    double wavefunc(int, int, double);
    void writelog(int,int);
    double **unclassical; int unclassicalind;
    double **E, **Ewkb, **Ewkb1, **norm, **I, **rmax;
    int loadwkbdata, loadqmdata, loadwkb1data;
  protected:
    double s;
    int nmax,mmin,mmax;
    int method;
    void initval(char*);
    int mi(int mlocal) {return -mmin+mlocal;}
    int nmi(int nlocal,int mlocal) {return nlocal*(mi(mmax)+1)+mi(mlocal);}
  public:
    Matching() {}
    Matching(char*,char*);
    ~Matching(){
      cerr << ":-) Good job!" << endl;
      delete Ewkb;
      delete Ewkb1;
      delete E;
      delete norm;delete rmax;
      cerr << "Matching finished!" << endl;
    }
    void Energy();
    int CheckError(int, int);
    int CorrectError();
    
    void sets(double s0) {s=s0;cerr << "s= " << s << endl;}
    void rwfile(double**&,char*,bool);
    double getE(int n, int m) {return E[n][mi(m)];}
    double getnorm(int n, int m) {return norm[n][mi(m)];}
    double getrmax(int n, int m) {return rmax[n][mi(m)];}
    void displayUC(ostream&);
};


// ********************** Matching applications ************************

class Statistics {
  private:
    char *qmfilename,*wkbfilename, *wkbfilename1,*Ifilename;
    int statnum,kkrit;
    int snum;
    int nmax,mmin,mmax;
    int loadqmdata, loadwkbdata, loadwkb1data;
    int saveqmdata, savewkbdata, savewkb1data;
    double **stN,**E1,***E;
    double smin, smax, ds;
  protected:
    void initval(char*);  
    int mi(int mlocal) {return -mmin+mlocal;}
    int nmi(int nlocal,int mlocal) {return nlocal*(mi(mmax)+1)+mi(mlocal);}
    int si(double s) {return (int)((s-smin)/ds);}
  public:
    Statistics(char*,char*);
    void GenerateE(char*, char*);
    void load(double ***&,char*);
    void stats();
    double nE(double);
    void display(ostream&, char*);
    void displayerror(ostream&, char*, char*);
    void displayerrorsmall(ostream&, char*, char*,char*);
    void displaymaslov(ostream&, char*);
    void displayenergy_maslov(int, double, ostream&, char*);
    void displayE_m(double, ostream&, char*);
    void displayI_m(double, ostream&, char*);
    void displayE_s(ostream&, char*);
    void displaymu_s(ostream&, char*);
    void displaymaslov_s(ostream&, char*,char*);
    void save(char*);
    ~Statistics(){ delete stN;delete E;
     cerr << "Statistics finished!" << endl;}
};

// **********************  Auxialarry ******************************

double bessj(int, double);

inline double besselj(int m, double x) {return (m >= 0) ? bessj(m,x) : ((m/2.0==m/2) ? bessj(-m,x) : -bessj(-m,x)); }
// J[-m] = (-1)^m * J[m]

inline double Theta(double x) {return x>=0 ? x : 0;}
inline double Heaviside(double x) {return x>=0 ? 1 : 0;}
inline int Sign(double a) {return a>=0 ? 1:-1;}
void displaymatrix(double**&, int,int,int,int,ostream&);
void graphf(double, double, int);
double interpolate(double,double*,double*,int);


// ********************** Roots of functions *****************************

/*
class rootsofbessel : public Dfunc {
private:
  int m;
public:
  rootsofbessel(int m1) {m=m1;Df=0;}
  double f(double x) {
    Df = besselj(m-1,x)-m/x*besselj(m,x);
    return besselj(m,x);}
};


class rootsofkummeru : public func {
private:
  double b,x;
public:
  rootsofkummeru(double x1, double b1) {x=x1;b=b1;}
  double f(double a);
};
*/


class Roots {
  public:
     double Bisect(func*,double,double,double,int&);
     double NRBisect(Dfunc*,double,double,double,int&);
     double NR(Dfunc*,double,double,int&);
     double Secant(func*,double,double,double,int&);
};

class integration {
  public:
    double Trapezoid(func*, double, double, int);
    double Midpoint(func*, double, double, int);
    double Gauss(func*, double, double, int);
    double Simpson(func*, double, double, int);
};

// *****   Class for the confluent hypergeometric functions **************
class parifuncs {
  public:
     parifuncs() {
       long d, prec =3;
       pari_init(60000000, 2);
       d = 19;
       if (d > 0) prec = (int)(d*pariK1+3);}
     double U(double a, double b, double z) {
       return gtodouble(hyperu(dbltor(a),dbltor(b),dbltor(z),prec));}
     double uu(double a, double b, double z) {
       return gtodouble(divrr(hyperu(dbltor(a),dbltor(b+1),dbltor(z),prec),
			 hyperu(dbltor(a),dbltor(b),dbltor(z),prec)));}
};


// ***************************   MAIN  **********************************
integration integrate;
parifuncs pf;
Roots rt;


int main(int argc, char *argv[]) {
  if ((argc!=3)&&(argc!=4)) {
      cerr << "args: inputfile outputfile [statfile]" << endl;
      return 0;}
  srand(time(0));

  // ********* Displaying the current *********    
/*
  Matching fi(argv[1],argv[2]);
  fi.Energy();
  int m,n; double s=5;
  current j;
  j.resetlog();
  long ltop=avma;
  for (n=0;n<=0;n++) {
      for (m=-4;m<=4;m++) {
	  current j(n, m, s , fi.getE(n,m), fi.getnorm(n,m), fi.getrmax(n,m));
//	  j.display(150,cout);
	  j.save(150);
      }
      avma=ltop;
  }
*/

  // ********* Obtaining states for fixed s *********    
//  Matching fi(argv[1],argv[2]);
//  fi.Energy();
//  fi.displayUC(cerr);


  // ********* Operations with a collection of eigenstates *********    
  Statistics st(argv[1],argv[2]);
  st.GenerateE(argv[1],argv[2]);
  st.save(argv[3]);

}



//   *********************************************************************
//   *********************************************************************
//   *********************************************************************


void displaymatrix(double**& M, int i0,int j0,int i1,int j1,ostream& os) {
  int iind,jind;
  for (iind=i0;iind<=i1;iind++) {
    for (jind=j0;jind<=j1;jind++)
      os << setw(10) << setprecision(5) << M[iind][jind];
    os << endl;
  }
}

void graphf(double xmin, double xmax, int N1) {
/*  int i;double x;
  int n=1;int m=1;int s=1;
	//findWKB0 WKB(n,m,s);
	//func* ptr = &WKB;
  for (i=0;i<=N1;i++) {
    x=xmin+(xmax-xmin)*i/(double)N1;
    cout << setw(15)<< setprecision(5) << x;
    for (m=-3;m<=3;m++)
      cout << setw(15)<< setprecision(5) << current(nmax,m,x);
			// Ehhez most definialni kell a matching objektumot !
	  //cout << setw(15)<< setprecision(5) << (ptr->f)(x);
    cout << endl;
    }*/
}


double interpolate(double x,double* xo,double* yo,int dim) {
  int i,j;
  double sum=0;
  double prod;
  for (i=0;i<=dim;i++) {
    prod = 1;
    for (j=0;j<=dim;j++)
       if (j!=i) prod *= (x-xo[j])/(xo[i]-xo[j]);
    sum += yo[i]*prod;
  }
  return sum;
}



//   *********************************************************************
//   *********************  Energy Statistics ****************************
//   *********************************************************************

Statistics::Statistics(char *paramfile, char *energyfile0)
{
  initval(paramfile);
  int i,j;


  qmfilename = new char[30]; wkbfilename = new char[30]; wkbfilename1 = new char[30]; Ifilename = new char[30]; 
  strcpy(qmfilename,energyfile0);
  strcpy(wkbfilename,energyfile0);
  strcpy(wkbfilename1,energyfile0);
  strcpy(Ifilename,energyfile0);
  strcat(qmfilename,".qm.dat");
  strcat(wkbfilename,".wkb.dat");
  strcat(wkbfilename1,".wkb1.dat");
  strcat(Ifilename,".I.dat");


  if (saveqmdata) {
      ofstream qmfile(qmfilename, ios::binary|ios::ate);  
      qmfile.close();
      ofstream Ifile(Ifilename, ios::binary|ios::ate);  
      Ifile.close();
  }
  if (savewkbdata) {
      ofstream wkbfile(wkbfilename, ios::binary|ios::ate);  
      wkbfile.close();
  }
  if (savewkb1data) {
      ofstream wkbfile1(wkbfilename1, ios::binary|ios::ate);  
      wkbfile1.close();
  }
  //The statistics files are reset if they are to be saved.
  

  E = new double**[snum+1]; 
  for (i=0;i<=snum;i++) 
    E[i] = new double*[nmax+1];
  for (i=0;i<=snum;i++) 
    for (j=0;j<=nmax;j++) 
      E[i][j] = new double[mi(mmax)+1];
  // allocats memory for E[s][n][m]

  E1 = new double*[nmax+1]; 
  for (i=0;i<=nmax;i++) E1[i] = new double[mi(mmax)+1];

  int kmax=nmi(nmax,mmax);kkrit=kmax;
  // WARNING: the maximum number of different s constrained
  stN = new double*[kmax+1]; 
  for (i=0;i<=kmax;i++) stN[i] = new double[snum+1];

  statnum=0;
}


void Statistics::initval(char *paramfile) {
  double number;char descr[30];
  ifstream infile(paramfile, ios::in);
  if (!infile) {
    cerr << "Error opening " << paramfile << "!" << endl;
    exit(0);}
  else  cout << endl << "Parameters read from " << paramfile << endl;
  loadqmdata=loadwkbdata=loadwkb1data=1;
  saveqmdata=savewkbdata=savewkb1data=1;
  while (infile >> descr >> number) {
    if (strcmp(descr,"mmin")==0) mmin=(int)number;
    if (strcmp(descr,"mmax")==0) mmax=(int)number;
    if (strcmp(descr,"nmax")==0) nmax=(int)number;
    if (strcmp(descr,"smin")==0) smin=number;
    if (strcmp(descr,"smax")==0) smax=number;
    if (strcmp(descr,"ds")==0) ds=number;
    if (strcmp(descr,"loadqmdata")==0) loadqmdata=(int)number;
    if (strcmp(descr,"loadwkbdata")==0) loadwkbdata=(int)number;
    if (strcmp(descr,"loadwkb1data")==0) loadwkb1data=(int)number;
    if (strcmp(descr,"saveqmdata")==0) saveqmdata=(int)number;
    if (strcmp(descr,"savewkbdata")==0) savewkbdata=(int)number;
    if (strcmp(descr,"savewkb1data")==0) savewkb1data=(int)number;
  }
  infile.close();
  if (loadwkb1data || loadqmdata) loadwkbdata=1;
  cout << "mmin = " << mmin << endl
       << "mmax = " << mmax << endl
       << "nmax = " << nmax << endl
       << "smin = " << smin << endl
       << "smax = " << smax << endl
       << "ds = " << ds << endl
       << "loadqmdata = " << loadqmdata << endl
       << "loadwkbdata = " << loadwkbdata << endl
       << "loadwkb1data = " << loadwkb1data << endl
       << "saveqmdata = " << saveqmdata << endl
       << "savewkbdata = " << savewkbdata << endl
       << "savewkb1data = " << savewkb1data << endl;
  snum = si(smax)+1;
}

void Statistics::GenerateE(char *paramfile,char *energyfile) {
  double s=smin;
  long ltop=avma;int i=0;
  if ((loadqmdata) && (loadwkbdata) && (loadwkb1data)) return;
  //The calculations need not be done if all data is loaded.
  while (s<=smax) {
    Matching psi(paramfile,energyfile);
    psi.sets(s);
    psi.Energy();
    s+=ds;
    if ((++i)%100==0) avma=ltop;
  }
}

void Statistics::load(double ***&M,char* file){
  double *stN0;
  int kmax=nmi(nmax,mmax);
  stN0 = new double[kmax+1];

  ifstream infile(file, ios::binary);
  if (!infile) {
    cerr << "Error opening " << file << "!" << endl;
    exit(0);
  }

  int n,m,i;i=0;
  while (!infile.eof()){
    infile.read((char*)stN0,(kmax+1)*sizeof(double));
    for (n=0;n<=nmax;n++)
      for (m=mmin;m<=mmax;m++) 
	M[i][n][mi(m)]=stN0[nmi(n,m)];
    i++;
  }
  i--;
  infile.close();
  cerr << i << " antidot solutions read from " << file << "." << endl; 
  cerr << "The database is "; if (i!=snum) cerr << "in"; 
  cerr << "complete!" << endl;
}

void Statistics::display(ostream& os, char* file) {
  double s=smin;int n,m; 
  load(E,file);
  while (s<=smax) {
    os << setw(10) << s;
    for (n=0;n<=nmax;n++)
      for (m=mmin;m<=mmax;m++)
	os << setw(15) << E[si(s)][n][mi(m)];
    os << endl;
    s+=ds;
  }
}

void Statistics::displayerror(ostream& os, char* file1, char* file2){
  int i,j;
  double ***EE;
  EE = new double**[snum+1]; 
  for (i=0;i<=snum;i++) 
    EE[i] = new double*[nmax+1];
  for (i=0;i<=snum;i++) 
    for (j=0;j<=nmax;j++) 
      EE[i][j] = new double[mi(mmax)+1];
  // allocats memory for EE[s][n][m]

  double s=smin; int n,m; 
  load(E,file1);
  load(EE,file2);
  while (s<=smax) {
    os << setw(10) << s;
    for (n=0;n<=nmax;n++)
      for (m=mmin;m<=mmax;m++)
	os << setw(15) << 
	  (E[si(s)][n][mi(m)]-EE[si(s)][n][mi(m)])/EE[si(s)][n][mi(m)];
    os << endl;
    s+=ds;
  }
}

void Statistics::displayerrorsmall(ostream& os, char* file1, char* file2, char* file3){
  int i,j;
  double ***EE,***EEE;
  EE = new double**[snum+1]; 
  for (i=0;i<=snum;i++) 
    EE[i] = new double*[nmax+1];
  for (i=0;i<=snum;i++) 
    for (j=0;j<=nmax;j++) 
      EE[i][j] = new double[mi(mmax)+1];
  EEE = new double**[snum+1]; 
  for (i=0;i<=snum;i++) 
    EEE[i] = new double*[nmax+1];
  for (i=0;i<=snum;i++) 
    for (j=0;j<=nmax;j++) 
      EEE[i][j] = new double[mi(mmax)+1];
  // allocats memory for EE[s][n][m], EEE[s][n][m]


  double s=smin; int n,m; 
  load(E,file1);
  load(EE,file2);
  load(EEE,file3);
  while (s<=smax) {
    os << setw(10) << s;
    n=0;m=4;
    os << setw(15) << 
	  (E[si(s)][n][mi(m)]-EE[si(s)][n][mi(m)])/EE[si(s)][n][mi(m)];
    os << setw(15) << 
	  (E[si(s)][n][mi(m)]-EEE[si(s)][n][mi(m)])/EEE[si(s)][n][mi(m)];
    os << endl;
    s+=ds;
  }
}

void Statistics::displaymaslov(ostream& os, char* file){
  double s=smin;int n,m; 
  load(E,file);
  while (s<=smax) {
    os << setw(10) << s;
    for (n=0;n<=nmax;n++)
      for (m=mmin;m<=mmax;m++){
	findWKB0 WKB(n,m,s);
	os << setw(15) << 2*s*WKB.nu(E[si(s)][n][mi(m)]/(2.*s),m/(2.*s))-n;
      }
    os << endl;
    s+=ds;
  }
}

void Statistics::displaymaslov_s(ostream& os, char* Efile, char* Wfile){
  double s=smin;int i,j,n,m;
  double mu,mu1,eps,eps1,b1,a,b,Delta,mas; 
  double ***EE;
  EE = new double**[snum+1]; 
  for (i=0;i<=snum;i++) 
    EE[i] = new double*[nmax+1];
  for (i=0;i<=snum;i++) 
    for (j=0;j<=nmax;j++) 
      EE[i][j] = new double[mi(mmax)+1];
  load(E,Efile);  load(EE,Wfile);
  n=0;
  while (s<=smax) {
    os << setw(15) << s;
    for (m=1;m<=8;m++) {
	findWKB0 WKB(n,m,s);
	os << setw(15) << 2*s*WKB.nu(E[si(s)][n][mi(m)]/(2.*s),m/(2.*s))-n;
	findWKB1 WKB1(n,m,s);

	eps=eps1=EE[si(s)][n][mi(m)]/(2.*s); mu=mu1=m/(2.*s);

	WKB1.nu(eps,mu,s,a,b,Delta); 
	a*=2*s; b*=2*s; Delta= Delta / (2.*s);
	if (eps>mu*mu) 
	    mas = (1 - 1/pi*
		atan((2./Delta + cos(2*b)) / (1 + sin(2*b))));
	else mas = 0.5;
	mas -= floor(mas); 
	os << setw(15) << mas;
    }
    os << endl;
  s+=ds;
  }
}


void Statistics::displayenergy_maslov(int m, double s, ostream& os, char* file){
  int n; 
  load(E,file);
  for (n=0;n<=nmax;n++){
    os << setw(15) << E[si(s)][n][mi(m)];
    findWKB0 WKB(n,m,s);
    os << setw(15) << 2*s*WKB.nu(E[si(s)][n][mi(m)]/(2.*s),m/(2.*s))-n;
    os << endl;
  }
}

void Statistics::displayE_m(double s, ostream& os, char* file){
  int n,m; 
  load(E,file);
  for (m=-12;m<=20;m++){
      os << setw(15) << m;
      for (n=0;n<=8;n++){
	  os << setw(15) << E[si(s)][n][mi(m)];
      }
      os << endl;
  }
}

void Statistics::displayI_m(double s, ostream& os, char* file){
  int n,m; 
  load(E,file);// I[s][n][m] is stored in the E array
  for (m=-17;m<=27;m++){
      os << setw(15) << m;
      for (n=0;n<=2;n++){
	  os << setw(15) << E[si(s)][n][mi(m)];
      }
      os << endl;
  }
}



void Statistics::displayE_s(ostream& os, char* file){
  int n,m; double s; 
  load(E,file);
  n=0;
  for (s=smin;s<=smax;s+=ds){
      os << setw(15) << s;
      for (m=-2;m<=2;m++){
	  os << setw(15) << E[si(s)][n][mi(m)];
      }
      os << endl;
  }
}

void Statistics::save(char *statfilename){
  char *fullstatfilename;
  fullstatfilename = new char[30];

  ofstream statfile;

  if (saveqmdata) {
      strcpy(fullstatfilename,statfilename);
      strcat(fullstatfilename,".qm.dat");
      statfile.open(fullstatfilename, ios::ate);
      display(statfile,qmfilename);
      statfile.close();
  }

  if (savewkbdata) {
      strcpy(fullstatfilename,statfilename);
      strcat(fullstatfilename,".wkb.dat");
      statfile.open(fullstatfilename, ios::ate);
      display(statfile,wkbfilename);
      statfile.close();
  }

  if (savewkb1data) {
      strcpy(fullstatfilename,statfilename);
      strcat(fullstatfilename,".wkb1.dat");
      statfile.open(fullstatfilename, ios::ate);
      display(statfile,wkbfilename1);
      statfile.close();
  }

  if (saveqmdata || savewkbdata) {
      strcpy(fullstatfilename,statfilename);
      strcat(fullstatfilename,".err.dat");
      statfile.open(fullstatfilename, ios::ate);
      displayerror(statfile,qmfilename,wkbfilename);
      statfile.close();
  }

  if (saveqmdata || savewkb1data) {
      strcpy(fullstatfilename,statfilename);
      strcat(fullstatfilename,".err1.dat");
      statfile.open(fullstatfilename, ios::ate);
      displayerror(statfile,qmfilename,wkbfilename1);
      statfile.close();
  }
 
  if (saveqmdata) {
      strcpy(fullstatfilename,statfilename);
      strcat(fullstatfilename,".mas.dat");
      statfile.open(fullstatfilename, ios::ate);
      displaymaslov(statfile,qmfilename);
      statfile.close();
  }

  /*
      strcpy(fullstatfilename,statfilename);
      strcat(fullstatfilename,".E_m.qm.dat");
      statfile.open(fullstatfilename, ios::ate);
      displayE_m(5,statfile,qmfilename);
      statfile.close();

      strcpy(fullstatfilename,statfilename);
      strcat(fullstatfilename,".E_m.wkb.dat");
      statfile.open(fullstatfilename, ios::ate);
      displayE_m(5,statfile,wkbfilename);
      statfile.close();


      strcpy(fullstatfilename,statfilename);
      strcat(fullstatfilename,".E_m.wkb1.dat");
      statfile.open(fullstatfilename, ios::ate);
      displayE_m(5,statfile,wkbfilename1);
      statfile.close();
*/

      strcpy(fullstatfilename,statfilename);
      strcat(fullstatfilename,".I.dat");
      statfile.open(fullstatfilename, ios::ate);
      display(statfile,Ifilename);
      statfile.close();


      strcpy(fullstatfilename,statfilename);
      strcat(fullstatfilename,".I_m.dat");
      statfile.open(fullstatfilename, ios::ate);
      displayI_m(5,statfile,Ifilename);
      statfile.close();



/*
      strcpy(fullstatfilename,statfilename);
      strcat(fullstatfilename,".E_s.qm.dat");
      statfile.open(fullstatfilename, ios::ate);
      displayE_s(statfile,qmfilename);
      statfile.close();

      strcpy(fullstatfilename,statfilename);
      strcat(fullstatfilename,".E_s.wkb.dat");
      statfile.open(fullstatfilename, ios::ate);
      displayE_s(statfile,wkbfilename);
      statfile.close();


      strcpy(fullstatfilename,statfilename);
      strcat(fullstatfilename,".E_s.wkb1.dat");
      statfile.open(fullstatfilename, ios::ate);
      displayE_s(statfile,wkbfilename1);
      statfile.close();
*/
/*
  strcpy(fullstatfilename,statfilename);
  strcat(fullstatfilename,".masenerg.dat");
  statfile.open(fullstatfilename, ios::ate);
  displayenergy_maslov(5,30,statfile,qmfilename);
  statfile.close();
*/

/*
  strcpy(fullstatfilename,statfilename);
  strcat(fullstatfilename,".mu_s.dat");
  statfile.open(fullstatfilename, ios::ate);
  displaymaslov_s(statfile,qmfilename,wkbfilename);
  statfile.close();
*/
}




double Statistics::nE(double e) {
  // how many states under a given energy bound
  int k0=0;int k1=nmi(nmax,mmax);
  while (k0!=k1-1) {
      if (stN[(k0+k1)/2][0]<e) k0=(k0+k1)/2;
      else k1=(k0+k1)/2;
  }
  return k0;
}

void Statistics::stats() {
  // how many states under a given energy bound
  // first sort the energy states in order
  int n,m,k,k1;
  double *stN0;
  int *stN1;
  int kmax=nmi(nmax,mmax);
  stN0 = new double[kmax+1];
  stN1 = new int[kmax+1];
  for (n=0;n<=nmax;n++)
    for (m=mmin;m<=mmax;m++) {
      stN0[nmi(n,m)]=E1[n][mi(m)];
    }
  for (k=0;k<=kmax;k++) stN1[k]=0;
  for (k=1;k<=kmax;k++)
    for (k1=0;k1<k;k1++) {
      if (stN0[k]<stN0[k1])  stN1[k1]++;
      else stN1[k]++;
    }
  for (k=0;k<=kmax;k++)
   	stN[stN1[k]][statnum] = stN0[k];

  if (kkrit > stN1[nmi(nmax,0)]) kkrit = stN1[nmi(nmax,0)];
  statnum++;
  //displaymatrix(E1,0,mi(mmin),nmax,mi(mmax),cerr);
}




//   *********************************************************************
//   ************************  Matching class ****************************
//   *********************************************************************

Matching::Matching(char *paramfile, char *output){
  initval(paramfile);
  int i;

  qmfilename = new char[30];
  wkbfilename = new char[30];
  wkbfilename1 = new char[30];
  normfilename = new char[30];
  Ifilename = new char[30];
  rmaxfilename = new char[30];
  logfilename = new char[30];

  strcpy(qmfilename,output);
  strcpy(wkbfilename,output);
  strcpy(wkbfilename1,output);
  strcpy(normfilename,output);
  strcpy(Ifilename,output);
  strcpy(rmaxfilename,output);
  strcpy(logfilename,output);
  
  strcat(qmfilename,".qm.dat");
  strcat(wkbfilename,".wkb.dat");
  strcat(wkbfilename1,".wkb1.dat");
  strcat(normfilename,".norm.dat");
  strcat(Ifilename,".I.dat");
  strcat(rmaxfilename,".rmax.dat");
  strcat(logfilename,".log");

  ofstream f(logfilename, ios::ate); f.close();

  E = new double*[nmax+1]; for (i=0;i<=nmax;i++) E[i] = new double[mi(mmax)+1];
  Ewkb = new double*[nmax+1]; for (i=0;i<=nmax;i++) Ewkb[i] = new double[mi(mmax)+1];
  Ewkb1 = new double*[nmax+1]; for (i=0;i<=nmax;i++) Ewkb1[i] = new double[mi(mmax)+1];
  norm = new double*[nmax+1]; for (i=0;i<=nmax;i++) norm[i] = new double[mi(mmax)+1];
  rmax = new double*[nmax+1]; for (i=0;i<=nmax;i++) rmax[i] = new double[mi(mmax)+1];
  I = new double*[nmax+1]; for (i=0;i<=nmax;i++) I[i] = new double[mi(mmax)+1];
  
  unclassicalind = 0;
  unclassical = new double*[2*mmax*nmax+1]; for (i=0;i<=2*mmax*nmax;i++) unclassical[i] = new double[3+1];
  // WARNING: the maximum number of unclassical records is approximated by 2*mmax*nmax
  
}


void Matching::initval(char *paramfile) {
  char descr[30];double number;  
  ifstream infile(paramfile, ios::in);
  if (!infile) {
    cerr << "Error opening " << paramfile << "!" << endl;
    exit(0);}
  else  cout << endl << "Parameters read from " << paramfile << endl;
  loadqmdata=loadwkbdata=loadwkb1data=1;
  while (infile >> descr >> number) {
    if (strcmp(descr,"s")==0) s=number;
    if (strcmp(descr,"mmin")==0) mmin=(int)number;
    if (strcmp(descr,"mmax")==0) mmax=(int)number;
    if (strcmp(descr,"nmax")==0) nmax=(int)number;
    if (strcmp(descr,"method")==0) method=(int)number;  
    if (strcmp(descr,"loadqmdata")==0) loadqmdata=(int)number;
    if (strcmp(descr,"loadwkbdata")==0) loadwkbdata=(int)number;
    if (strcmp(descr,"loadwkb1data")==0) loadwkb1data=(int)number;
    //    if (strcmp(descr,"xmin")==0) xmin=number;
    //    if (strcmp(descr,"xmax")==0) xmax=number;
    //    if (strcmp(descr,"N")==0) N=(int)number; Ezek a graph(xmin,xmax,N)-hez kellenek
  }
  infile.close();
  if ((!loadqmdata) || (!loadwkb1data)) loadwkbdata=0;
  // The wkb data is recalculated if needed by qm or wkb1.
  cout << "s = " << s << endl
       << "mmin = " << mmin << endl
       << "mmax = " << mmax << endl
       << "nmax = " << nmax << endl
       << "method = " << method << endl
       << "loadqmdata = " << loadqmdata << endl
       << "loadwkbdata = " << loadwkbdata << endl
       << "loadwkb1data = " << loadwkb1data << endl
       << endl;
}



void Matching::BuildWKB() {
try {
  int n,m,Code;
  func* ptr;
  cerr << "Calculating WKB energies: " << endl;
  for (n=0;n<=nmax;n++) {
    for (m=mmin;m<=mmax;m++) {
      findWKB0 WKB(n,m,s);
      ptr = &WKB;
      Ewkb[n][mi(m)]=rt.Bisect(ptr,0.0001,2*n+2*Theta(m)+1+1e-5,1e-10,Code);
      if (Code==0) cerr <<  '.'; else cerr <<  '!';
    }
  }
  cerr << endl;}
catch(...) {
  cerr << "WKB fucked up!" << endl;
  throw;};
}

void Matching::BuildWKB1() {
try {
  int n,m,Code;
  func* ptr;
  cerr << "Calculating WKB1 energies: " << endl;
  double xL,xR;
  for (n=0;n<=nmax;n++) {
    for (m=mmin;m<=mmax;m++) {
      findWKB1 WKB(n,m,s);
      ptr = &WKB;
      xL = Ewkb[n][mi(m)] - 10*T;
      xR = xL + 10*T;
      if ( Ewkb[n][mi(m)] > m*m/(2.*s)) 
	  Ewkb1[n][mi(m)]= rt.Secant(ptr,xL,xR,T,Code);
      else Ewkb1[n][mi(m)]= Ewkb[n][mi(m)];
      if (Code==0) cerr <<  '.'; else cerr <<  '!';
    }
  }
  cerr << endl;}
catch(...) {
  cerr << "WKB1 fucked up!" << endl;
  throw;};
}



int Matching::CheckError(int n, int m) {
  if ((fabs(1-Ewkb[n][mi(m)]/E[n][mi(m)])>0.2)
      ||(fabs(Ewkb[n][mi(m)]-E[n][mi(m)])>0.25)
      ||((m!=mmin)&&(m<=0)&&(E[n][mi(m-1)]!=-1)&&(E[n][mi(m-1)]-E[n][mi(m)]<0.001))
      ||((m!=mmax)&&(m>0)&&(E[n][mi(m)]!=-1)&&(E[n][mi(m)]-E[n][mi(m-1)]<0.001))
      ||((n!=0)&&(E[n-1][mi(m)]!=-1)&&(E[n][mi(m)]-E[n-1][mi(m)]<0.001))
      ) {
 		unclassical[unclassicalind][0]=n;
 		unclassical[unclassicalind][1]=s;
 		unclassical[unclassicalind][2]=m;
 		unclassicalind++;
 		return -4;
 	}
	return 0;
}

int Matching::CorrectError() {
  int u,i,j,n,m;
  double s;
  double guess[4],bound[4],result[4];
  double d,dd,ddd,xL,xR;int Code;
  double y[3],x[3];
  int** UC;
  UC = new int*[nmax+1]; 
  for (i=0;i<=nmax;i++) UC[i] = new int[mi(mmax)+1];
  for (u=0;u<unclassicalind;u++) 
    UC[(int)unclassical[u][0]][mi((int)unclassical[u][2])]=1;

  for (u=unclassicalind-1;u>=0;u--) {
    n=(int)unclassical[u][0]; 
    s=unclassical[u][1];
    m=(int)unclassical[u][2];
    cerr << n  << "," << m << " | ";
    if ((m!=mmax)&&(fabs((E[n][mi(m)]-E[n][mi(m+1)]) - 
			 (Ewkb[n][mi(m)]-Ewkb[n][mi(m+1)]) )<0.05)) {
      if ((n!=nmax)&&(UC[n+1][mi(m)]==0)) {
	if (fabs((E[n][mi(m)]-E[n+1][mi(m)]) - 
	      (Ewkb[n][mi(m)]-Ewkb[n+1][mi(m)]) )<0.05) {
	  cerr << "This is correct" << endl;
	  UC[n][mi(m)]=0;
	}
      }
      else {
	cerr << "This is correct" << endl;
	UC[n][mi(m)]=0;
      }
    }
    else if ((n!=nmax)&&(UC[n+1][mi(m)]==0)) {
	if ((fabs((E[n][mi(m)]-E[n+1][mi(m)]) - 
	      (Ewkb[n][mi(m)]-Ewkb[n+1][mi(m)]) )<0.05)) {
	  cerr << "This is correct" << endl;
	  UC[n][mi(m)]=0;
	}
      }
  }

  j=0;
  for (u=0;u<unclassicalind;u++) {
    n=(int)unclassical[u][0]; 
    s=unclassical[u][1];
    m=(int)unclassical[u][2];
    if (UC[n][mi(m)]==0) j++;
    else {
      unclassical[u-j][0]=unclassical[u][0];
      unclassical[u-j][1]=unclassical[u][1];
      unclassical[u-j][2]=unclassical[u][2];
    }
  }
  unclassicalind -= j;


  //******************************************************************
  //*******************  Correcting the errors  **********************

  for (u=0;u<unclassicalind;u++) {

    n=(int)unclassical[u][0];
    s=unclassical[u][1];
    m=(int)unclassical[u][2];
    for (i=0;i<=3;i++ ){guess[i]=bound[i]=result[i]=0;}
    bound[3]=1000;

    cerr<< endl;
    cerr << "(" << n <<"," << m << ")" << endl;
    //**************  The low m extrapolation *************************
    j=m;
    for (i=0;((i<3)&&(j>mmin));i++) {
      j--;
      while ((UC[n][mi(j)]==1)&&(j>mmin)) j--;
      cerr << "g0: " << j << " ";
      if (i==0) {
	if ((j>=0)||(m<=0)) bound[0]=E[n][mi(j)];
	else {
	  if (UC[n][mi(0)]==0) bound[0]=E[n][mi(0)];
	  else bound[0]=0;
	}
      }
      y[i]=E[n][mi(j)]; x[i]=j;
    }//The enery decreases with m if m<0, and increases if m>0.
    i--;
    if ((bound[0]!=0)&&(i>0)) guess[0]=interpolate(m,x,y,i);
    if (j==-mmax) {
      while ((UC[n][mi(j)]==1)&&(j<0)) j++;     
      if (j<0) bound[0]= E[n][mi(j)] + 2*(Ewkb[n][mi(-mmax)]-Ewkb[n][mi(j)]);
    }


    
    //**************  The high m extrapolation *************************
    j=m;
    for (i=0;((i<3)&&(j<mmax));i++) {
      j++;
      while ((UC[n][mi(j)]==1)&&(j<mmax)) j++;
      cerr << "g1: " << j << " ";
      if (i==0) {
	if ((j<=0)||(m>=0)) bound[1]=E[n][mi(j)];
	else {
	  if (UC[n][mi(0)]==0) bound[1]=E[n][mi(0)];
	  else bound[1]=0;
	}
      }//The enery decreases with m if m<0, and increases if m>0.
      y[i]=E[n][mi(j)];x[i]=j;
    }
    i--;
    if ((bound[1]!=0)&&(i>0)) guess[1]=interpolate(m,x,y,i);
    if (j==mmax){
      while ((UC[n][mi(j)]==1)&&(j>0)) j--;
      if (j>0) bound[1]= E[n][mi(j)] + 2*(Ewkb[n][mi(j)]-Ewkb[n][mi(j-1)]);
    }

    //**************  The low n extrapolation *************************    
    j=n;
    for (i=0;((i<3)&&(j>0));i++) {
      j--;
      while ((UC[j][mi(m)]==1)&&(j>0)) j--;
      cerr << "g2: " << j << " ";
      if ((i==0)&&(UC[j][mi(m)]==0)) bound[2]=E[j][mi(m)];
      //The enery is monotonic in n.
      y[i]=E[j][mi(m)]; x[i]=j;
    }
    i--;
    if (i>0) guess[2]=interpolate(n,x,y,i);
     

    //**************  The high n extrapolation *************************
    j=n;
    for (i=0;((i<3)&&(j<nmax));i++) {
      j++;
      while ((UC[j][mi(m)]==1)&&(j<nmax)) j++;
      cerr << "g3: " << j << " ";
      if ((i==0)&&(UC[j][mi(m)]==0)) bound[3]=E[j][mi(m)];
    //The enery is monotonic in n.
      y[i]=E[j][mi(m)];x[i]=j;
    }
    i--;
    if (i>0) guess[3]=interpolate(n,x,y,i);

    //*************  manipulating the bounds and guesses **************
    cerr << "guesses= " << guess[0] << " " << guess[1] << " " 
	 << guess[2] << " " << guess[3] << endl;

    if (bound[0]>bound[1]) {
      d=bound[0];
      bound[0]=bound[1];
      bound[1]=d;
    }//assign 0 index to the smaller and 1 to the larger bound
    if (m==0) {
      bound[1]=bound[0];
      bound[0]=0;
    }//The n=const states assume the minimum energy at m=0.
    for (i=0;i<=3;i++)
      if ((guess[i]<bound[0])||(guess[i]>bound[1])||
	  (guess[i]<bound[2])||(guess[i]>bound[3]))
	guess[i]=0;//guess is out of bounds
    if ((bound[1]-bound[0])/(bound[3]-bound[2])<0.33) {
      guess[2]=guess[3]=0;
    }
    else if ((bound[1]-bound[0])/(bound[3]-bound[2])>3) {
      guess[0]=guess[1]=0;
    }//eliminates irrelevant guesses

    //*************  Correcting the unclassical states ****************
    if ((n!=nmax)&&(n!=0))
      if ((UC[n-1][mi(m)]==0)&&(UC[n+1][mi(m)]==0))
	if ((E[n][mi(m)]-bound[2]>0.001)&&(bound[3]-E[n][mi(m)]>0.001)){
	  UC[n][mi(m)]=0;
	  cerr<< endl << "good state1! (" << n <<"," << m << ")" <<endl;
	}//surely good state if it is between neighboring states
    if (n==0)
      if (UC[n+1][mi(m)]==0)
	if (bound[3]-E[n][mi(m)]>0.001){
	  UC[n][mi(m)]=0;
	  cerr<< endl << "good state2! (" << n <<"," << m << ")";
	}//surely good state if it is between neighboring states	
    if (((n!=0)&&(bound[1]-bound[0] < 0.75*(Ewkb[n][mi(m)]-Ewkb[n-1][mi(m)])))||
	((n==0)&&(bound[1]-bound[0] < 0.75*(Ewkb[n+1][mi(m)]-Ewkb[n][mi(m)]))))
      if ((E[n][mi(m)]-bound[0]>0)&&(bound[1]-E[n][mi(m)]>0)){
	UC[n][mi(m)]=0;
	cerr<< endl << "good state3! (" << n <<"," << m << ")" << endl;
      }//surely a good state if it is between neighboring states

    d=1000;
    for (i=0;i<=3;i++) 
      if (guess[i]!=0) {
	xL=sqrt(guess[i]*(1.- T*0.01));
	xR=sqrt(guess[i]*(1.- T*0.001));
	result[i] = rootfinder(xL,xR,Code,m);
	result[i] = result[i]*result[i];
	if (d>(fabs(result[i]-guess[i]))) {
	  d=fabs(result[i]-guess[i]);
	  E[n][mi(m)]=result[i];
	}
      }
    
    cerr << "bounds= " <<bound[0] << " " << bound[1] << " " 
	 << bound[2] << " " << bound[3] << endl;
    cerr << "guesses= " << guess[0] << " " << guess[1] << " " 
	 << guess[2] << " " << guess[3] << endl;
    cerr << "results= " << result[0] << " " << result[1] << " " 
	 << result[2] << " " << result[3] << endl;

    //************ If still bad: probabilistic method ***************
    if ((E[n][mi(m)]-max(bound[0],bound[2])< 0.001)||
	(min(bound[1],bound[3])-E[n][mi(m)]< 0.001)) {
      d=1000;dd=ddd=0;
      for (i=0;i<=3;i++)
	for (j=0;j<=3;j++)
	  if ((guess[i]!=0)&&(guess[j]!=0)&&
	      (guess[i]-guess[j]>0)&&(d>guess[i]-guess[j])) {
	    d = guess[i]-guess[j];
	    dd = guess[i];
	    ddd  = guess[j];
	  }//the guess values with the least difference is chosen 
      if (dd==0)
	for (i=0;i<=3;i++) if (guess[i]!=0) dd=guess[i];
      if ((ddd==0)||(dd-ddd<0.05))
	ddd = (max(bound[0],bound[2])+min(bound[1],bound[3]))/2.0;
      //the mean value of the bounds is used if only one guess exists
      if (dd-ddd<0.05) {
	if (n!=nmax) d = (Ewkb[n+1][mi(m)]-Ewkb[n][mi(m)])/30.0;
	else d = (Ewkb[n][mi(m)]-Ewkb[n-1][mi(m)])/30.0;
	dd += d;
	ddd -= d;
      }//excluding the possibility that the guess values are equal;
		       
      cerr << "guesses between: "<< setw(10) << dd << setw(10) << ddd<< endl;
      i=0;//Executing a probabilistic search
      while (((E[n][mi(m)]-max(bound[0],bound[2])< 0.001)||
	     (min(bound[1],bound[3])-E[n][mi(m)]< 0.001))&&(i<100)) {
	xR = rand()/(RAND_MAX+1.0);
	d = ddd + xR*(dd-ddd);
	xL=sqrt(d*(1.- T*0.01));
	xR=sqrt(d*(1.- T*0.001));
	E[n][mi(m)] = rootfinder(xL,xR,Code,m);
	E[n][mi(m)] = E[n][mi(m)]*E[n][mi(m)];
	cerr << setw(10) << d << setw(10) << 	E[n][mi(m)] << endl;
	i++;
      }
      cerr << "State corrected in steps: " << i << endl; 
    }

    cerr << "final= " << E[n][mi(m)] << endl; 
    writelog(n,m);
  }

  j=0;
  for (u=0;u<unclassicalind;u++) {
    n=(int)unclassical[u][0]; 
    s=unclassical[u][1];
    m=(int)unclassical[u][2];
    if (UC[n][mi(m)]==0) j++;
    else {
      unclassical[u-j][0]=unclassical[u][0];
      unclassical[u-j][1]=unclassical[u][1];
      unclassical[u-j][2]=unclassical[u][2];
    }
  }
  unclassicalind -= j;

  cerr << "TOK JO2!!!" << endl;

  return unclassicalind;
}




void Matching::BuildExact() {
  int n,m,Code;
  double xL,xR,x,T1;
  cerr << "Calculating the exact energies: " << endl;
  for (n=0;n<=nmax;n++) {
    for (m=mmin;m<=mmax;m++) {
      if (sqrt(Ewkb[n][mi(m)]*2*s)+10>abs(m)){
	if (m!=0) {
	  xL=sqrt(Ewkb[n][mi(m)]*(1.- T*0.01));
	  xR=sqrt(Ewkb[n][mi(m)]*(1.- T*0.001));
	  x = rootfinder(xL,xR,Code,m);
	}
	else {
	  T1=T; T=1e-4;
	  if (m>0) {
	    xL=sqrt(Ewkb1[n][mi(m)]*(1.- T*0.1));
	    xR=sqrt(Ewkb1[n][mi(m)]*(1.- T*0.01));
	  }
	  else {
	    xL=sqrt(Ewkb[n][mi(m)]*(1.- T*0.1));
	    xR=sqrt(Ewkb[n][mi(m)]*(1.- T*0.01));
	  }
	  x =rootfinder(xL,xR,Code,m);
	  T = T1;
	}
	E[n][mi(m)] = x*x;
	writelog(n,m);
      	if (Code==0) CheckError(n,m);
	else {
	  unclassical[unclassicalind][0]=n;
	  unclassical[unclassicalind][1]=s;
	  unclassical[unclassicalind][2]=m;
	  unclassicalind++;
	}
      }
      else {E[n][mi(m)] = Ewkb[n][mi(m)];Code = -3;}
      switch (Code){
       case 0  : cerr << '.'; break;
       case -3 : cerr << '_'; break;
       case -4 : cerr << '+'; break;
       default: cerr << '!' << Code; break;
      }
    }
  }
  strcat(logfilename,".1"); 
  x=CorrectError();
  cerr << "Residue: " << x << endl;

  unclassicalind=0;
  for (n=0;n<=nmax;n++) 
    for (m=mmin;m<=mmax;m++) 
      if (E[n][mi(m)]!=-1) CheckError(n,m);

  logfilename[strlen(logfilename)-1]='2'; 
  x=CorrectError();
  cerr << "Residue: " << x << endl;
}

void Matching::BuildNorm() {
	int n,m;
	func *ptr;
  cerr << "Calculating the wave function normalizations: " << endl;
  for (n=0;n<=nmax;n++) {
    for (m=mmin;m<=mmax;m++) {
      rmax[n][mi(m)]= sqrt(E[n][mi(m)]-(m-s)+fabs(m-s)) +
	              sqrt(E[n][mi(m)]-(m-s)-fabs(m-s));
       //sqrt(E[n][mi(m)])+sqrt( E[n][mi(m)]-2*(m-s) ); 06.22.2004
      unnormalized hf0(m,s,E[n][mi(m)],true);
      ptr = &hf0;
      norm[n][mi(m)] = sqrt(integrate.Simpson(ptr,0,1.5*rmax[n][mi(m)],100));
      current cf(n, m, s , E[n][mi(m)], norm[n][mi(m)], rmax[n][mi(m)],true);
      ptr = &cf;
      I[n][mi(m)] = integrate.Simpson(ptr,0.0001,1.5*rmax[n][mi(m)],100);
      cerr << n<< "  " << m << " " << s << " rmax= " << rmax[n][mi(m)] 
	   << " I= " << I[n][mi(m)] << endl;
//      cerr << '.';
    }
  }
  cerr << endl;
}


void Matching::Energy() {
  try {
    if (!loadwkbdata) {
	BuildWKB();
	rwfile(Ewkb,wkbfilename,true);
    }
    if (!loadwkb1data) {
	BuildWKB1();
	rwfile(Ewkb1,wkbfilename1,true);
    }
    if (!loadqmdata) {
      BuildExact();
      rwfile(E,qmfilename,true);
    }
    if (!loadqmdata) {
      BuildNorm();
      rwfile(norm,normfilename,true);
      rwfile(I,Ifilename,true);
      rwfile(rmax,rmaxfilename,true);
    }
  }
  catch(...) {
    cerr << "Shit! Shit! Shit! Energy levels fucked up!" << endl;
    throw;
  };
}

//   *********************************************************************


void Matching::displayUC(ostream& os) {
  int i,k;
  os << "Number of unclassical eigenstates is: " << unclassicalind << endl;
  if (unclassicalind>0) {
  	os << setw(5) << 'n' << setw(5) << 's' << setw(5) << 'm' << endl;
  	for (k=0;k<unclassicalind;k++) {
	  for (i=0;i<=2;i++) os << setw(5) << unclassical[k][i];
	  os << setw(10) << Ewkb[(int)unclassical[k][0]][mi((int)unclassical[k][2])];
	  os << setw(10) << Ewkb1[(int)unclassical[k][0]][mi((int)unclassical[k][2])];
	  os << setw(10) << E[(int)unclassical[k][0]][mi((int)unclassical[k][2])];
	  os << endl;
  	}
  }
}


void Matching::rwfile(double **&M,char* file,bool w){
  double *stN0;
  int kmax=nmi(nmax,mmax);
  stN0 = new double[kmax+1];
  int n,m;
  if (w) {
    for (n=0;n<=nmax;n++)
      for (m=mmin;m<=mmax;m++) 
      	stN0[nmi(n,m)]=M[n][mi(m)];
    ofstream outfile(file, ios::binary|ios::app);
    if (!outfile) {
      cerr << "Error opening " << file << "!" << endl;
      exit(0);
    }
    outfile.write((char*)stN0,(kmax+1)*sizeof(double));
    outfile.close();
    cerr << "File written to " << file  << "." << endl;
  }
  else {
    ifstream infile(file, ios::binary);
    if (!infile) {
      cerr << "Error opening " << file << "!" << endl;
      exit(0);
    }
    infile.read((char*)stN0,(kmax+1)*sizeof(double));
    infile.close();
    cerr << "File read from " << file << "." << endl; 
    for (n=0;n<=nmax;n++)
      for (m=mmin;m<=mmax;m++) 
      	M[n][mi(m)]=stN0[nmi(n,m)];
  }
}

void Matching::writelog(int n,int m) {
  double hiba;
  ofstream logfile(logfilename, ios::app);
 	if (!logfile) {
  		cerr << "Error opening " << logfilename << "!" << endl;
    	exit(0);
  }
  hiba = 1-E[n][mi(m)]/Ewkb[n][mi(m)];
  logfile << setw(10) << n << setw(10) << s << setw(10) << m
	  << setw(10) << setprecision(5) << Ewkb[n][mi(m)]
	  << setw(10) << setprecision(5) << Ewkb1[n][mi(m)]
	  << setw(10) << setprecision(5) << E[n][mi(m)]
	  << setw(15) << setprecision(5) << hiba;
  if ((fabs(hiba)>0.2)||(fabs(Ewkb[n][mi(m)]-E[n][mi(m)])>0.5))
    logfile << " Warning!";
  logfile << endl;
  logfile.close();
}



//   *********************************************************************

//   *********************************************************************
double Matching::wavefunc(int n1, int m1, double r)
{
  unnormalized psi(m1,s,E[n1][mi(m1)]);
  return psi.f(r)/norm[n1][mi(m1)];
}


double Matching::rootfinder(double xL, double xR,int& Code, int m) {
  matchDE dfo(m,s);
  matchE fo(m,s);
  func* ptr= &fo;
  Dfunc* Dptr = &dfo;
  double x; 
  switch (method) {
      case 0: x = rt.Bisect(ptr,xL,xR,T,Code); break;
      case 1: x = rt.NRBisect(Dptr,xL,xR,T,Code); break;
      case 2: x = rt.NR(Dptr,xL,T,Code);       break;
      case 3: x = rt.Secant(ptr,xL,xR,T,Code); break;
    default:  cerr << "Which method? (0-3)" << endl; return 0; break;
 }
  return x;
}



//   *********************************************************************
//   *********************************************************************
//   *********************************************************************

// Roots of the first argument
//double rootsofkummeru::f(double a) {
//  return pf.U(a,b,x);
//}

double matchDE::duE(double a, double b, double z) {
  //  ( (U(a+da,b+1)/U(a+da,b) - U(a,b+1)/U(a,b)) * U(a+da,b)/U(a,b)) )/da
      GEN A = dbltor(a); GEN B = dbltor(b);  GEN BB= dbltor(b+1);
      GEN Z = dbltor(z); GEN TT = dbltor(T);
      GEN UB=hyperu(A,B,Z,prec);
      GEN UPB=hyperu(gadd(A,TT),B,Z,prec);
      return gtodouble(divrr(
          mulrr(subrr(divrr(hyperu(gadd(A,TT),BB,Z,prec),UPB),
		      divrr(hyperu(A,BB,Z,prec),UB)),
                divrr(UPB,UB)),TT));}

/*
Elvileg tort alakban felirva csak a szamlalo zerus helyeit kene keresni, de ezzel az a
baj, hogy tipikusan kurvanagy/kurvanagy kifejezesek vannak. Tulcsordalasveszely stb. Csak 
kis energiakra lehet hasznos a szamlalo zerushelyeinek keresese. A wkb gyokok jok, ezzel
eleg dolgozni nagy energiakon.
Az alabbi 3 megoldas egyiket kell csak meghagyni.
*/


//Az alapeset amikor a torteket hasznaljuk.
/*
double matchDE::f(double x) {
  double b = 1 + fabs(m-s);
  double c = -1./2. - (fabs(m-s) + abs(m))/(2.0*s);
  double a = Theta(m-s) + (1-x*x)/2;
  double JpJ=besselj(abs(m)-1,x*sqrt(2*s))/besselj(abs(m),x*sqrt(2*s));
  Df = -x*(1.+JpJ*JpJ) + (abs(2*m)-1.+1./sqrt(2.*s))*JpJ - x*duE(a,b,s);
  return x/sqrt(2*s)*JpJ + pf.uu(a,b,s) + c;}

double matchE::f(double x) {
  double b = 1 + fabs(m-s);
  double c = -1./2. - (fabs(m-s)+m)/(2.*s);
  double a = Theta(m-s) + (1.-x*x)/2.;
  return x/sqrt(2*s)*besselj(m-1,x*sqrt(2*s))/besselj(m,x*sqrt(2*s))
    +c+pf.uu(a,b,s);
}
*/


//A bessel szingularitasokat eltuntetjuk

double matchDE::f(double x) {
  double b = 1 + fabs(m-s);
  double c = -1./2. - (fabs(m-s) + abs(m))/(2.*s);
  double a = Theta(m-s) + (1-x*x)/2;
  double cuu = sqrt(2*s)*(c+pf.uu(a,b,s));
  Df = besselj(abs(m)-1,x*sqrt(2*s))*(1/sqrt(2*s)+(abs(m)-1)+cuu)
    - besselj(abs(m),x*sqrt(2*s))*(x + abs(m)/x*cuu + x*duE(a,b,s));
  return x/sqrt(2*s)*besselj(abs(m)-1,x*sqrt(2*s))+
    besselj(abs(m),x*sqrt(2*s))*c + pf.uu(a,b,s)*besselj(abs(m),x*sqrt(2*s));
}

/*
double matchE::f(double x) {
  double b = 1 + fabs(m-s);
  double c = -1./2. - (fabs(m-s)+m)/(2.*s);
  double a = Theta(m-s) + (1.-x*x)/2.;
  return x/sqrt(2*s)*besselj(m-1,x*sqrt(2*s))+
    besselj(m,x*sqrt(2*s))*c + pf.uu(a,b,s)*besselj(m,x*sqrt(2*s));
}
*/
//Baszd meg itt nem konzisztens az m es abs(m)!



// A bessel es a Kummer szingularitasokat is megszuntetjuk (ehhez nem csinaltam derivaltat)

double matchE::f(double x) {
  double b = 1 + fabs(m-s);
  double c = -1./2. - (fabs(m-s)+m)/(2.*s);
  double a = Theta(m-s) + (1.-x*x)/2.;
  return (x/sqrt(2*s)*besselj(m-1,x*sqrt(2*s))+
    besselj(m,x*sqrt(2*s))*c)*pf.U(a,b,s) + pf.U(a,b+1,s)*besselj(m,x*sqrt(2*s));
}


//   *********************************************************************
//   *********************************************************************
//   *********************************************************************

double findWKB0::nu(double eps,double mu) {
  double nuph=0;
  if (eps>=mu*mu) nuph = sqrt(eps - mu*mu) -
		 Sign(mu-1)*(eps-2*mu+1)*acos( fabs(mu-1)/sqrt(eps-2*mu+1) ) +
		(Sign(mu)*eps - 2*fabs(mu))*acos(fabs(mu)/sqrt(eps));
  if (mu==0) nuph=sqrt(eps) + pi/2.*eps + (eps+1)*atan(sqrt(eps));
  nuph/=2*pi;
  if (mu<0) nuph+=0.5*eps;
  if (mu>=1) nuph+=0.5*(eps-2*mu+1);
  return nuph ;
}

double findWKB0::f(double E) {
    return 2*s*nu(E/(2.*s),m/(2.*s))-0.5-n;
}
//   *********************************************************************

void findWKB1::nu(double eps,double mu, double s, 
		    double &a, double &b, double &delta) {
  double de;
  double ONE=1./(4*s*s);
  de=eps-(mu*mu); //-ONE); ------------------> trukkozeshez
  delta=fabs(mu)/(2.*pow(fabs(de),1.5));
  if (mu==0) {
    a =sqrt(eps); 
    b =pi/2.*eps + (eps+1)*atan(sqrt(eps));
  }
  else {
    if (de>=0) {
      if (de>eps) cerr << "Error: mu*mu<ONE" << endl; 
      /*
      double mu0=Sign(mu)*sqrt(mu*mu-ONE);
      double muP= mu-0.5;
      double muP1=Sign(muP)*sqrt(muP*muP -ONE);
      double eps1=eps + muP1 - muP;
      double mu1= muP1 + 0.5;*/ //---------------------> trukkozeshez
      double mu0=mu;
      double mu1=mu;
      double eps1=eps;
      if (mu*mu<1.1*ONE) {
	  mu0=mu1=mu;
	eps1=eps;
	delta=fabs(mu)/(2.*pow(fabs(eps-mu*mu),1.5));
      }
      a= 2*sqrt(eps - mu0*mu0) - 2*fabs(mu0)*acos(fabs(mu0)/sqrt(eps));
      b= - sqrt(eps1 - mu1*mu1) 
	- Sign(mu1-1)*(eps1-2*mu1+1)*acos( fabs(mu1-1)/sqrt(eps1-2*mu1+1) ) 
	+ Sign(mu1)*eps1*acos(fabs(mu1)/sqrt(eps1))
	+ Heaviside(-mu1)*pi*eps1 
	+ Heaviside(mu1-1)*pi*(eps1-2*mu1+1);
    }
    else {
      delta = 0;
      a = 0;
      b = Heaviside(-mu)*pi*eps + Heaviside(mu-1)*pi*(eps-2*mu+1);
    }
  }
/*  double b1;
	if (!((0<b)&&(b<50))) {
      double mu0=mu;
      double mu1=mu-0.5;
      double eps1=eps;
      b1= - sqrt(eps1 - mu1*mu1) ;
	- Sign(mu1-1)*(eps1-2*mu1+1)*acos( fabs(mu1-1)/sqrt(eps1-2*mu1+1) ) 
	+ Sign(mu1)*eps1*acos(fabs(mu1)/sqrt(eps1))
	+ Heaviside(-mu1)*pi*eps1 
	+ Heaviside(mu1-1)*pi*(eps1-2*mu1+1);
	    cerr << setw(3) << n
		 << setw(3) << m
		 << setw(7) << s
		 << setw(15) << b
		 << setw(15) << eps1 - mu1*mu1
		 << setw(15) << b1<< endl;}*/

  a /= 2;
  b /= 2;
}



double findWKB1::f(double E) {
  double a,b,Delta,ans;
  nu(E/(2.*s),m/(2.*s),s,a,b,Delta);
  //tan(2*s*a - pi/4  ) = tan( -2*s*b + pi/4) + Delta/(2*s)
  Delta = Delta / (2.*s);
  a *= 2*s;
  b *= 2*s;
  //return tan(2*s*a - pi/4  ) - tan( -2*s*b + pi/4) - Delta/(2*s);
  //return a+b - pi*(n+0.5) - atan(Delta/2.*(1 + cos(a-b)/sin(a+b)));
  ans = a+b - ( pi*n + pi - atan( (2./Delta + cos(2*b))
				   /(1 + sin(2*b))       ));
  //if (m>0) ans += pi;
  return ans;
}


// itt x=sqrt(E)
double unnormalized::W(double a, double b, double z) {
  GEN Z = dbltor(z); GEN MNEGY = dbltor(-4.); GEN KETTO = dbltor(2.);
  GEN MCS= dbltor((double(m)-s)); GEN X=divrr(mulrr(Z,Z),KETTO);
  GEN ppp= gpow(Z,gabs(MCS,prec),prec);
  return gtodouble(mulrr(mulrr(gexp(divrr(mulrr(Z,Z),MNEGY),prec),ppp),hyperu(dbltor(a),dbltor(b),X,prec) ));
  //  return exp(-z*z/4)*pow(z,fabs(m-s))*U(a,b,z*z/2);
}

double unnormalized::f(double r)
  //E=E[n][mi(m)]
{
  double b = 1 + fabs(m-s);
  double a = Theta(m-s) + (1-E)/2.;
  double CC= besselj(abs(m),sqrt(2*s*E))/W(a,b,sqrt(2*s));
  double y= (r<=sqrt(2*s)) ? besselj(abs(m),sqrt(E)*r) : CC*W(a,b,r);
  return  squared ? y*y*2*pi*r : y;
}



double current::f(double r) {
  unnormalized psi(m,s,E,true);//This is 2*pi*r*psi^2
  return 
      (2.*m + 2.*(r*r/2-s)*Heaviside(r-sqrt(2*s)))*
      (psi.f(r)/(2*pi*r)) / (norm*norm* ((normalbasis) ? r : 1));
/*    unnormalized psi(m,s,E); 
      return psi.f(r)/norm;*/
}

void current::display(int K, ostream& os){
    double x=0; 
    int i; double dx;
    for (i=0;i<=K;i++) {
	dx = 1.3*rmax/K ; // The current is displayed over the maximum cl. rad.
	if ((n==0)&&(abs(m)<=1)&&(s==5))dx = 6.0/(double(K)); //CAUTION THIS IS ONLY FOR m=-1,0,1; n=0; s=5

	x += dx;
	os << setw(15) << setprecision(5) << x;
	os << setw(15) << setprecision(5) << f(x);
	os << endl;
    }
}

void current::save(int K){
  currentfilename = new char[30]; 
  char index[30];
  strcpy(currentfilename,"current.");
  gcvt(n,2,index); strcat(currentfilename,index); strcat(currentfilename,".");
  gcvt(m,2,index); strcat(currentfilename,index); strcat(currentfilename,".");
  gcvt(s,2,index); strcat(currentfilename,index); 

  char datfilename[30];
  strcpy(datfilename,currentfilename);
  strcat(datfilename,".dat");
  ofstream currentfile;
  currentfile.open(datfilename, ios::ate);
  display(K,currentfile);
  currentfile.close();
  cerr << "File written to " << datfilename << endl;
  writelog();
}

void current::writelog() {
  char figfilename[30];
  strcpy(figfilename,currentfilename);
  strcat(figfilename,".fig");

  char figfilename1[30];
  strcpy(figfilename1,currentfilename);
  strcat(figfilename1,"b.fig");


  char datfilename[30];
  strcpy(datfilename,currentfilename);
  strcat(datfilename,".dat");

  ofstream logfile(logfilename, ios::app);
  if (!logfile) {
      cerr << "Error opening " << logfilename << "!" << endl;
      exit(0);
  }
  logfile << setw(15) << setprecision(5) << E
          << setw(10) << m << setw(10) << s 
	  << setw(30) << datfilename
	  << setw(30) << figfilename
	  << setw(30) << figfilename1
	  << setw(15) << setprecision(5) << rmax
	  << 5 << 8 << endl;
  logfile.close();
}

void current::resetlog(){
  ofstream logfile(logfilename, ios::ate);
  if (!logfile) {
      cerr << "Error creating " << logfilename << "!" << endl;
      exit(0);
  }
  logfile.close();
}
//   *********************************************************************
//   ******************  Root finding algorithms *************************
//   *********************************************************************

double Roots::Bisect(func* pf,
	     double xL, double xR, 
	     double T, int& Code) {
  try {	
    double yL=pf->f(xL);
    double y=pf->f(xR);
    double D0=1e-3;double D=D0;
    if (yL*y>0) {cerr << endl << yL << "   " << y << "  boundary!!" <<endl; return 0;}
    while (yL*y>0) {
      xL=sqrt(xL*xL-D*D);
      yL=pf->f(xL);
      D+=D0;
    };
    /*        cout << "Using the Bisect method" << endl
	 << "Left  estimate  xL :  " << xL << endl
         << "Right estimate  xR :  " << xR << endl
	 << "Tolerance        T :  " << T << endl    
	 << setw(5)  << "n" << setw(14) << "xL"
	 << setw(14) << "xR" << setw(14) << "f(xL)" 
	 << setw(14) << "f(xR)" << setw(14) << "dx" << endl;  */
    double x;
    int n=0;
    //    long ltop=avma;
    do {
      x = (xL+xR)/(double)2.0;
      /*        cout << setw(5) << n 
	   << setw(14)<< setprecision(5) << xL
	   << setw(14)<< setprecision(5) << xR
	   << setw(14)<< setprecision(5) << f(xL)
	   << setw(14)<< setprecision(5) << f(xR)
	   << setw(14)<< setprecision(5) << x-xL << endl;      */
      y = pf->f(x); yL= pf->f(xL);
      //      if ((int)n/10.0==n/10.0) avma=ltop;
      Sign(y) == Sign(yL) ? xL = x : xR = x;
    } while (++n < 1000 && fabs(xR-xL) >= T); 
    if (fabs(xR-xL) < T) fabs(yL)<1000 ? Code=0 : Code=-2;
      else Code=-1;
    return x;}
  catch(...) {
    cerr << "\nwhile executing bisect ; now xL,xR,T = " 
         << xL << ',' << xR << ',' << T << endl;
    throw;
  }
}

//osszeakadt a PARI-val, ezert ki kellett cserelni az abs()-okat fabs()-ra
// es a template<>-et kihuztam
double Roots::NR(Dfunc* FdF,
	  double x, double T, int& Code) {
  try {
    /*    cout << "Using the Newton-Raphson method" << endl
	 << "Initial estimate  x :  " << x << endl
	 << "Tolerance         T :  " << T << endl    
	 << setw(5) << "n" << setw(10) << "x" << setw(10) << "f(x)"
	 << setw(10)<< "f'(x)" << setw(10) << "dx" << endl;*/
    int n=0;
    double Fx,dx;
    //long ltop=avma;
    do {
        Fx = FdF->f(x);
	/*	cout << setw(5) << n << setw(10) << setprecision(2) << x
             << setw(10)<< setprecision(2) << Fx
	     << setw(10)<< setprecision(2) << dF
	     << setw(10)<< setprecision(2) << dx << endl;*/
	if (Fx == 0) dx=0;
	  else dx = Fx/(FdF->Df);
	//	if ((int)n/10.0==n/10.0) avma=ltop;
	x -= dx; } 
      while (++n < 1000 && fabs(dx) >= T);
    if (fabs(dx) < T) fabs(Fx)<10000 ? Code=0 : Code=-2;
       else Code=-1;
    return x; }
  catch(...) {
    cerr << "\nwhile executing NR ; now x,T = " << x << ',' << T;
    throw; }}

double Roots::NRBisect(Dfunc* FdF,
		double xL, double xR,
		double T, int& Code) {
  try {
    char Done;
    int n=0;
    double x,h,yL,y;
    yL=FdF->f(xL);
    y=FdF->f(xR);
    double D0=1e-3;double D=D0;
    while (yL*y>0) {
      xL=sqrt(xL*xL-D*D);
      yL=FdF->f(xL);
      D+=D0;
    } ;
    /*cout << "Using the hybrid Newton-Raphson bisection method" << endl
	 << "Left  estimate  xL :  " << xL << ",  y= " << yL << endl
         << "Right estimate  xR :  " << xR << ",  y= " << y  << endl
	 << "Tolerance        T :  " << T << endl    
	 << setw(5)  << "n" << setw(10) << "xL" << setw(10) << "x"
	 << setw(10) << "xR" << setw(10) << "f(x)" 
	 << setw(10)<< "f'(x)" << setw(10) << "dx" << endl;  */
    x = (xL + xR)/2;
    h = x - xL;
    yL = FdF->f(xL);
    //    long ltop=avma;
    do {
        y = FdF->f(x);
	Done = (h < T);
	if (!Done) {
	  /*cout << setw(5) << n 
	       << setw(10)<< setprecision(2) << xL
	       << setw(10)<< setprecision(2) << x
	       << setw(10)<< setprecision(2) << xR
	       << setw(10)<< setprecision(2) << y
	       << setw(10)<< setprecision(2) << yp
	       << setw(10)<< setprecision(2) << h << endl;*/
	  if (Sign(yL) == Sign(y)) {
	      xL = x;  yL =y; }
	    else 
	      xR = x;
	  h = (y == 0 ? 0 : y/(FdF->Df));
	  x -= h;
	  h = fabs(h);
	  //	  if ((int)n/10.0==n/10.0) avma=ltop;
	  if (x < xL || xR < x) {
	    x = (xL + xR)/2;
	    h = x - xL; }}}
      while (n++ < 1000 && !Done);
    fabs(y) >= 1000 ? Code=-2 : Code = -!Done;
    return x; }
  catch(...) {
    cerr << "\nwhile executing NRBisect ; now xL,xR,T = " 
         << xL << ',' << xR << ',' << T << endl;
    throw; }}

double Roots::Secant(func* fp,double x0,double x1,
		double T,int& Code) {
    try {
      /*cout << "Using the Secant method" << endl
	 << "Left  estimate  x0 :  " << x0 << endl
         << "Right estimate  x1 :  " << x1 << endl
	 << "Tolerance        T :  " << T << endl    
	 << setw(5)  << "n" << setw(10) << "x0"
	 << setw(10) << "x1" << setw(10) << "f(x0)" 
	 << setw(10) << "f(x1)" << setw(10) << "dx" << endl;  */
      int    n  = 0;
      double dx = x1 - x0;
      double y0 = fp->f(x0);
      double y1=fp->f(x1);double dy;
      //      long ltop=avma;
      do {
	/*cout << setw(5) << n 
	       << setw(10)<< setprecision(2) << x0
	       << setw(10)<< setprecision(2) << x1
	       << setw(10)<< setprecision(2) << y0
	       << setw(10)<< setprecision(2) << y1
	       << setw(10)<< setprecision(2) << dx << endl;*/
	  y1 = fp->f(x1);
	  dy = y1 - y0;
	  dx = -y1*dx/dy;
	  x0 = x1; y0 = y1;
	  //	  if ((int)(n/10.0)==n/10.0) avma=ltop;
	  x1 += dx; 
      } while (++n < 1000 && fabs(dx) >= T);
      //      avma=ltop;
      if (fabs(dx) < T) fabs(y1)<1000 ? Code=0 : Code=-2;
      else Code=-1;
      return x1; }
    catch(...) {
      cerr << "\nwhile executing Secant ; now x0,x1,T = "
	   << x0 << ',' << x1 << ',' << T << endl;
      throw; }}



//   *********************************************************************
//   *******************  Integration algorithms *************************
//   *********************************************************************



double integration::Trapezoid(func* fp, double a, double b, int n)
// n applications of basic Trtapezoid Rule interval [a, b]
{
 double h = (b - a)/n;
 double s = 0.5*( fp->f(a) + fp->f(b) );
 double x = a;
 for (int i=1; i<n; ++i)
 {
  x += h;
  s += fp->f(x);
 }
 return h*s;
}

double integration::Midpoint(func* fp, double a, double b, int n)
// n applications of basic Midpoint Rule over interval [a,b]
{
 double h = (b - a)/n;
 double x = a + 0.5*h;
 double s = fp->f(x);
 for (int i=1; i<n; ++i)
 {
  x += h;
  s += fp->f(x);
 }
 return h*s;
} 

double integration::Gauss(func* fp, double a, double b, int n)
// n applications of Gauss Rule over interval [a,b]
// 2n  subintervals
{
 double r = 1.0/sqrt(3.);
 double H = (b - a)/(2.0*n);
 double v = r*H;
 double s = 0.0;
 double u = a + H;
 for (int i=0; i<n; ++i)
 {
  s +=fp->f(u - v) + fp->f(u + v);
  u += 2*H;
 }
 return H*s;
}
 
double integration::Simpson(func* fp, double a, double b, int n)
// n application of Simpson's Rule over interval [a,b]
// 2n subintervals
{
 double H = (b - a)/n;
 double t = fp->f(a + 0.5*H);
 double s = 0.5*( fp->f(a) + fp->f(b) );
 double x = a;
 double y = a + 0.5*H;
 for (int i=1; i<n; ++i)
 {
  x += H;
  y += H;
  s += fp->f(x);
  t += fp->f(y);
 } 
 s += 2.0*t;
 return (H*s)/3.0;
}













